/****************************************************************************************
*
*   Disclaimer   This software code and all associated documentation, comments or other
*  of Warranty:  information (collectively "Software") is provided "AS IS" without
*                warranty of any kind. MICRON TECHNOLOGY, INC. ("MTI") EXPRESSLY
*                DISCLAIMS ALL WARRANTIES EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
*                TO, NONINFRINGEMENT OF THIRD PARTY RIGHTS, AND ANY IMPLIED WARRANTIES
*                OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE. MTI DOES NOT
*                WARRANT THAT THE SOFTWARE WILL MEET YOUR REQUIREMENTS, OR THAT THE
*                OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE.
*                FURTHERMORE, MTI DOES NOT MAKE ANY REPRESENTATIONS REGARDING THE USE OR
*                THE RESULTS OF THE USE OF THE SOFTWARE IN TERMS OF ITS CORRECTNESS,
*                ACCURACY, RELIABILITY, OR OTHERWISE. THE ENTIRE RISK ARISING OUT OF USE
*                OR PERFORMANCE OF THE SOFTWARE REMAINS WITH YOU. IN NO EVENT SHALL MTI,
*                ITS AFFILIATED COMPANIES OR THEIR SUPPLIERS BE LIABLE FOR ANY DIRECT,
*                INDIRECT, CONSEQUENTIAL, INCIDENTAL, OR SPECIAL DAMAGES (INCLUDING,
*                WITHOUT LIMITATION, DAMAGES FOR LOSS OF PROFITS, BUSINESS INTERRUPTION,
*                OR LOSS OF INFORMATION) ARISING OUT OF YOUR USE OF OR INABILITY TO USE
*                THE SOFTWARE, EVEN IF MTI HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
*                DAMAGES. Because some jurisdictions prohibit the exclusion or
*                limitation of liability for consequential or incidental damages, the
*                above limitation may not apply to you.
*
*                Copyright 2003 Micron Technology, Inc. All rights reserved.
*
****************************************************************************************/

`timescale 1ps/1ps

module statetable( 
    input UTYPE_dutconfig dutconfig,
    output UTYPE_dutmodeconfig modeconfig,
    output UTYPE_dutstate dutstate
		   );

   // default params
   parameter MAX_DQ_BITS = 36;
   parameter MAX_QK_BITS = 4;

   import rldram3_package::*;

   // init cycle number and cycle time
   parameter INITCYCLENUM = 1000;
   parameter INITCYCLETIME = 1000000;
   
   // save data per cycle
   typedef   struct 	  packed {
    time     cycletime;
				  } UTYPE_cyclestatedata;

   parameter MAXCYCLEPIPE = 10000;  // value > the biggest cycle based spec
   UTYPE_cyclestatedata cycledata[MAXCYCLEPIPE-1:0];

   function void clearcycledata(output UTYPE_cyclestatedata cycledata);
      begin
	 cycledata.cycletime=0;
      end
   endfunction

   // timing params
   UTYPE_sharedtiming timing;

   // check cycle or check time
   typedef enum {CKTM, CKCY} UTYPE_checktype;

   // cycle check variables
   int cWR[MAX_BANKS-1:0][1:0]  , cLastWRaddr[MAX_BANKS-1:0][1:0], cRD[MAX_BANKS-1:0][1:0], cRT;
   int cWR1[MAX_BANKS-1:0][1:0]  , cLastWRaddr1[MAX_BANKS-1:0][1:0];
   int cWR2[MAX_BANKS-1:0][1:0]  , cLastWRaddr2[MAX_BANKS-1:0][1:0];
   int cLMR, cLMRDLLRESET, cLastRD, cLastWR;
   int cREF1[MAX_BANKS-1:0][1:0], cREFM[MAX_BANKS-1:0][1:0];
   int cRdTrain, cMMD[1:0];
   int cScanE, cScanX, scanPipeCntr;
   reg LastWRrankbit, LastRDrankbit;
    
   // current time
   time 		     currenttime, currenttck;
   int 					 currentcycle, prevcycle;				 

   // mode reg
   reg [MAX_MR_BITS-1:0]    mr0, mr1, mr2;

   // array data
   reg [MAX_DQ_BITS*MAX_BURST_LEN-1:0] arraydata[int];

   // last access time for each array location
   time 			       arrayaccesstime[int];

   // TB flag
   reg 				       TB;
   
   parameter 			       DEBUG_TRAINING = 0;

   // lastcmdpkt from updatetable, used in mux mode
   UTYPE_cmdpacket lastcmdpkt;

   // TAP
   UTYPE_taptype nextTAPstate;
   // assign 			       dutstate.curTAPstate = curTAPstate;
   // assign 			       dutstate.curTAPIRstate = curTAPIRstate;
   
   bit [INST_LEN-1:0] 		       TAPIR, TAPIRcap;
   bit 				       TAPBYPASS;
   bit [ID_LEN-1:0] 		       TAPID;
   bit [BSR_LEN-1:0] 		       TAPBSR, TAPBSRcap, TAPBSRupdate;
   bit [BSR_LEN_EXT-1:0]	       TAPBSREXT, TAPBSREXTcap, TAPBSREXTupdate;

   //external chip state during i/o capture
   struct 			       packed {
    reg 			       clk;
    reg 			       clk_;
    reg 			       dm;
    reg [1:0] 			       dk;
    reg [1:0] 			       dk_;
    reg [2:0] 			       bank;
    reg [21:0] 			       addr;
    reg 			       cs_;
    reg 			       we_;
    reg 			       ref_;
    reg 			       zq;
    reg [35:0] 			       dq;
					       } ext_chip_state;

   //internal chip state as shifted from BSR
   struct 			       packed {
    reg 			       clk;
    reg 			       clk_;
    reg 			       dm;
    reg [1:0] 			       dk;
    reg [1:0] 			       dk_;
    reg [1:0] 			       qk;
    reg [1:0] 			       qk_;
    reg 			       qvld;
    reg [2:0] 			       bank;
    reg [21:0] 			       addr;
    reg 			       cs_;
    reg 			       we_;
    reg 			       ref_;
    reg 			       zq;
    reg [35:0] 			       dq;
    reg [35:0] 			       dq_en;
					       } int_chip_state;
   // timing functions
   function int minnops;
      input 			       UTYPE_cmdpacket cmdpkt;
      output [40*8-1:0] 	       cmdmsg;				 
      begin: querytable_func
         UTYPE_cmdtype cmd;
         int bank, bankl, addr;
         int tCK;
         int nops, WL, RL, BL, targetcycle, i, j;
	 reg rank;
         cmd = cmdpkt.cmd;
         bank = cmdpkt.bank & dutconfig.bankaddrmask;
         addr = cmdpkt.addr & dutconfig.addrmask;

	 if (dutconfig.ddpranked==1) 
	   rank = addr[dutconfig.numaddr-1];
	 else
	   rank = 0;
	 
         tCK = cmdpkt.tCK;
         nops=0;

	 RL=_RL();
	 WL=_WL();
         BL = modeconfig.BL;
         cmdmsg = "";
         
         // does targetcycle meet minimun timing
         targetcycle = currentcycle;
         
         case (cmd)
           cmdWR, cmdWRx: begin
              nops = cmdAllcheckspec(cmdmsg,tCK, targetcycle, nops);
	      for(i=0;i<numbankWRITE(modeconfig.WRITE);i++) begin
                 bankl = banklWRITE(modeconfig.WRITE,i,bank);
		 nops = checkspec1(tCK, targetcycle, cWR[bankl][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "WR to WR tRC check"); 
		 nops = checkspec1(tCK, targetcycle, cRD[bankl][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "RD to WR tRC check"); 
		 nops = checkspec1(tCK, targetcycle, cWR[bankl][rank], CKCY, modeconfig.BL/2+1, nops, cmdmsg, "WR to WR samebank BL/2+1 check"); 
		 nops = checkspec1(tCK, targetcycle, cRD[bankl][rank], CKCY, modeconfig.BL/2, nops, cmdmsg, "RD to WR samebank BL/2 check"); 
		 nops = checkspec1(tCK, targetcycle, cREF1[bankl][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "REF1 to WR tRC check"); 
		 nops = checkspec1(tCK, targetcycle, cREFM[bankl][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "REFM to WR tRC check");
	      end
	      nops = checkspec1(tCK, targetcycle, cLastWR, CKCY, modeconfig.BL/2, nops, cmdmsg, "WR to WR BL/2 check"); 
              if (dutconfig.sio&&dutconfig.archtype==RLDRAM3)
                addr=addr; // no check
              else if(dutconfig.sio&&dutconfig.archtype==RLDRAM3DDP&&dutconfig.ddpranked==1) begin
 	          if (modeconfig.AM && cmdpkt.addryflag===1'b0) begin
		      // model case, no check
		      addr = addr;
	          end
  	          else begin
		      // testbench use
                      if (addr[dutconfig.numaddr-1]!==LastRDrankbit)
	                nops = checkspec1(tCK, targetcycle, cLastRD, CKCY, modeconfig.BL/2, nops, cmdmsg, "RD to WR BL/2 check"); 
                  end
              end
              else
	        nops = checkspec1(tCK, targetcycle, cLastRD, CKCY, modeconfig.BL/2, nops, cmdmsg, "RD to WR BL/2 check"); 
           end
           
           cmdRT, cmdRTx: begin
              nops = cmdAllcheckspec(cmdmsg,tCK, targetcycle, nops);
	      nops = checkspec1(tCK, targetcycle, cRT, CKCY, 2, nops, cmdmsg, "RT to RT DataBurst check"); 
              nops = checkspec1(tCK, targetcycle, cLMRDLLRESET, CKCY, timing.tDLLkc, nops, cmdmsg, "DLLRESET to RT tDLLk check"); 	
           end

           cmdRD, cmdRDx: begin
              nops = cmdAllcheckspec(cmdmsg,tCK, targetcycle, nops);
              nops = checkspec1(tCK, targetcycle, cRD[bank][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "RD to RD tRC check"); 
              nops = checkspec1(tCK, targetcycle, cRD[bank][rank], CKCY, modeconfig.BL/2+1, nops, cmdmsg, "RD to RD samebank BL/2+1 check"); 
              nops = checkspec1(tCK, targetcycle, cREF1[bank][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "REF1 to RD tRC check"); 
              nops = checkspec1(tCK, targetcycle, cREFM[bank][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "REFM to RD tRC check"); 
              nops = checkspec1(tCK, targetcycle, cLMRDLLRESET, CKCY, timing.tDLLkc, nops, cmdmsg, "DLLRESET to RD tDLLk check"); 	
	      nops = checkspec1(tCK, targetcycle, cLastRD, CKCY, modeconfig.BL/2, nops, cmdmsg, "RD to RD BL/2 check"); 
              if (dutconfig.sio&&dutconfig.archtype==RLDRAM3)
                addr=addr; // no check
              else if(dutconfig.sio&&dutconfig.archtype==RLDRAM3DDP&&dutconfig.ddpranked==1) begin
 	          if (modeconfig.AM && cmdpkt.addryflag===1'b0) begin
		      // model case, no check
		      addr = addr;
	          end
  	          else begin
		      // testbench use
                      if (addr[dutconfig.numaddr-1]!==LastWRrankbit)
	                nops = checkspec1(tCK, targetcycle, cLastWR, CKCY, modeconfig.BL/2+1, nops, cmdmsg, "WR to RD BL/2+1 check"); 
                  end
              end
              else
	        nops = checkspec1(tCK, targetcycle, cLastWR, CKCY, modeconfig.BL/2+1, nops, cmdmsg, "WR to RD BL/2+1 check"); 
	      // always check diff addr
	      nops = checkspec1(tCK, targetcycle, cWR[bank][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "WR to RD tRC check"); 
              nops = checkspec1(tCK, targetcycle, cWR[bank][rank], CKCY, modeconfig.BL/2+2, nops, cmdmsg, "WR to RD samebank BL/2+2 check"); 
	      if (modeconfig.AM && cmdpkt.addryflag===1'b0) begin
		 // model case
		 addr = addr;
	      end
	      else begin
		 // testbench use
		 if (modeconfig.AM) 
		   MuxXY2A(cmdpkt.addr, cmdpkt.addry, addr);
		 addr = addr & dutstate.bladdrmask;
		 if (addr==cLastWRaddr[bank][rank]) // same addr wr to rd
		   nops = checkspec1(tCK, targetcycle, cWR[bank][rank], CKCY, WL+BL/2, nops, cmdmsg, "WR to RD tWTR_sameaddr check"); 
		 else if (addr==cLastWRaddr1[bank][rank]) // same addr wr to rd
		   nops = checkspec1(tCK, targetcycle, cWR1[bank][rank], CKCY, WL+BL/2, nops, cmdmsg, "WR to RD tWTR_sameaddr check"); 
		 else if (addr==cLastWRaddr2[bank][rank]) // same addr wr to rd
		   nops = checkspec1(tCK, targetcycle, cWR2[bank][rank], CKCY, WL+BL/2, nops, cmdmsg, "WR to RD tWTR_sameaddr check"); 
	      end
           end

	   cmdRDy: begin
	      // model use
	      if (modeconfig.AM && cmdpkt.addryflag===1'b0) begin
		 MuxXY2A(lastcmdpkt.addr, cmdpkt.addr, addr);
		 addr = addr & dutstate.bladdrmask;
		 if (addr==cLastWRaddr[bank][rank]) // same addr wr to rd
		   nops = checkspec1(tCK, targetcycle, cWR[bank][rank], CKCY, WL+BL/2+1, nops, cmdmsg, "WR to RD tWTR_sameaddr check"); 
		 else if (addr==cLastWRaddr1[bank][rank]) // same addr wr to rd
		   nops = checkspec1(tCK, targetcycle, cWR1[bank][rank], CKCY, WL+BL/2+1, nops, cmdmsg, "WR to RD tWTR_sameaddr check"); 
		 else if (addr==cLastWRaddr2[bank][rank]) // same addr wr to rd
		   nops = checkspec1(tCK, targetcycle, cWR2[bank][rank], CKCY, WL+BL/2+1, nops, cmdmsg, "WR to RD tWTR_sameaddr check"); 
	      end
	   end
	   
           cmdLMR, cmdLMRx: begin
              nops = cmdAllcheckspec(cmdmsg,tCK, targetcycle, nops);
              nops = checkspec1(tCK, targetcycle, cLastWR, CKCY, modeconfig.tRC, nops, cmdmsg, "WR to LMR tRC check"); 
              nops = checkspec1(tCK, targetcycle, cLastRD, CKCY, modeconfig.tRC, nops, cmdmsg, "RD to LMR tRC check"); 
              nops = checkspec1(tCK, targetcycle, cLastWR, CKCY, WL+BL/2+1, nops, cmdmsg, "WR to LMR WL+BL/2+1 check"); 
              nops = checkspec1(tCK, targetcycle, cLastRD, CKCY, RL+BL/2+1, nops, cmdmsg, "RD to LMR RL+BL/2+1 check"); 
              nops = checkspec1(tCK, targetcycle, cRT, CKCY, 8+1, nops, cmdmsg, "RT to LMR DataBurst check"); 
	      for (i=0; i< dutconfig.numbank; i++)  begin
		 nops = checkspec1(tCK, targetcycle, cREF1[i][0], CKCY, modeconfig.tRC, nops, cmdmsg, "REF1 to LMR tRC check"); 
		 nops = checkspec1(tCK, targetcycle, cREFM[i][0], CKCY, modeconfig.tRC, nops, cmdmsg, "REFM to LMR tRC check");
		 if (dutconfig.ddpranked==1)  begin	
		    nops = checkspec1(tCK, targetcycle, cREF1[i][1], CKCY, modeconfig.tRC, nops, cmdmsg, "REF1 to LMR tRC check"); 
		    nops = checkspec1(tCK, targetcycle, cREFM[i][1], CKCY, modeconfig.tRC, nops, cmdmsg, "REFM to LMR tRC check");
		 end
	      end
           end
           
           cmdREF1, cmdREF1x, cmdREFM, cmdREFMx: begin
              nops = cmdAllcheckspec(cmdmsg,tCK, targetcycle, nops);
	      
	      if (modeconfig.REF && modeconfig.AM && cmdpkt.addryflag===1'b0) begin
		 // model case
		 addr = addr;
	      end
	      else begin
		 // testbench use
		 if(modeconfig.REF) begin
		    if (modeconfig.AM) begin
		       MuxXY2A(cmdpkt.addr, cmdpkt.addry, addr);
		       addr = addr & dutconfig.addrmask;
		    end
		    j=0;
		    for (i=0; i< dutconfig.numbank; i++) 
		      if (addr[i]) begin
			 nops = checkspec1(tCK, targetcycle, cWR[i][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "WR to REF tRC check"); 
			 nops = checkspec1(tCK, targetcycle, cRD[i][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "RD to REF tRC check");
			 nops = checkspec1(tCK, targetcycle, cREF1[i][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "REF1 to REF tRC check"); 
			 nops = checkspec1(tCK, targetcycle, cREFM[i][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "REFM to REF tRC check"); 
			 if (j==1) 
			   nops = checkspec1(tCK, targetcycle, cMMD[rank], CKCY, timing.tMMDc, nops, cmdmsg, "tMMD to REF Multi-Bank check"); 
			 j=1;
		      end
		 end
		 else begin
		    nops = checkspec1(tCK, targetcycle, cWR[bank][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "WR to REF tRC check"); 
		    nops = checkspec1(tCK, targetcycle, cRD[bank][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "RD to REF tRC check");
		    nops = checkspec1(tCK, targetcycle, cREF1[bank][rank], CKCY, modeconfig.tRC, nops, cmdmsg, "REF1 to REF tRC check"); 
		    nops = checkspec1(tCK, targetcycle, cREFM[bank][rank], CKTM, modeconfig.tRC, nops, cmdmsg, "REFM to REF tRC check"); 
		 end
	      end
           end

	   cmdREFMy:
	     // model use
	     if (modeconfig.REF && modeconfig.AM && cmdpkt.addryflag===1'b0) begin
		MuxXY2A(lastcmdpkt.addr, cmdpkt.addr, addr);
		addr = addr & dutconfig.addrmask;
		for (i=0; i< dutconfig.numbank; i++) 
		  if (addr[i]) begin
		     nops = checkspec1(tCK, targetcycle, cWR[i][rank], CKCY, modeconfig.tRC+1, nops, cmdmsg, "WR to REFM tRC check"); 
		     nops = checkspec1(tCK, targetcycle, cRD[i][rank], CKCY, modeconfig.tRC+1, nops, cmdmsg, "RD to REFM tRC check");
		     nops = checkspec1(tCK, targetcycle, cREF1[i][rank], CKCY, modeconfig.tRC+1, nops, cmdmsg, "REF1 to REFM tRC check"); 
		     nops = checkspec1(tCK, targetcycle, cREFM[i][rank], CKCY, modeconfig.tRC+1, nops, cmdmsg, "REFM to REFM tRC check"); 
		  end
	     end
         endcase
	 
	 minnops = nops;
      end
   endfunction

   // common timing checks
   function int cmdAllcheckspec(
    inout [20*20-1:0] 	 cmdmsg,
    input 		 int tCK, 
    input 		 int targetcycle, 
    input 		 int nopsin);
      begin
	 int nops;
	 nops = nopsin;
	 nops = checkspec1(tCK, targetcycle, cLMR, CKCY, timing.tMRSCc, nops, cmdmsg, "ALLCMD to LMR tMRSC check"); 
	 cmdAllcheckspec = nops;
      end
   endfunction // int

   // update state table
   function void updatetable (
    reg   GenExpData,
    input UTYPE_cmdpacket cmdpkt,
    input int cmdextranops
			      );
      begin: updatetable_func
	 UTYPE_cmdtype cmd, covcmd;
	 int bank, bankl;
	 int addr;
	 int nops, WL, BL, RL, cycle, i,j;
	 int tCK;
	 reg [20*20-1:0] 	 cmdmsg;				 
	 reg 			 rank;
	 cmd = cmdpkt.cmd;
	 covcmd = cmd;
	 bank = cmdpkt.bank & dutconfig.bankaddrmask;
         addr = cmdpkt.addr & dutconfig.addrmask;
	 
	 if (dutconfig.ddpranked==1) 
	   rank = addr[dutconfig.numaddr-1];
	 else
	   rank = 0;

	 tCK = cmdpkt.tCK;
	 nops=0;
	 cycle=currentcycle;
	 RL=_RL();
	 WL=_WL();
	 BL = modeconfig.BL;

	 // dutstate defaults
	 dutstate.dly1=1*tCK;
	 dutstate.dly2=2*tCK;
	 dutstate.dly3=3*tCK;
	 dutstate.activebank='0;
	 dutstate.refbank='0;
	 
	 case (cmd)
	   cmdINIT: begin
	      inittable();
	   end
	   
	   cmdWR, cmdWRy: begin
 	      if (modeconfig.AM && cmd == cmdWRy) begin
		 MuxXY2A(lastcmdpkt.addr, cmdpkt.addr, addr);
		 bank = lastcmdpkt.bank & dutconfig.bankaddrmask;
	         if (dutconfig.ddpranked==1) 
	           rank = lastcmdpkt.addr[dutconfig.numaddr-1];
	      end
  	      addr = addr & dutstate.bladdrmask;

	      if (cmd == cmdWRy) 
		cycle=prevcycle;

              if (numbankWRITE(modeconfig.WRITE)==4)
		cMMD[rank] = cycle;
	      for(i=0;i<numbankWRITE(modeconfig.WRITE);i++) begin
                 bankl = banklWRITE(modeconfig.WRITE,i,bank);

		 // save previous 2
		 cWR2[bankl][rank] = cWR1[bankl][rank];
		 cLastWRaddr2[bankl][rank]=cLastWRaddr1[bankl][rank];
		 cWR1[bankl][rank] = cWR[bankl][rank];
		 cLastWRaddr1[bankl][rank]=cLastWRaddr[bankl][rank];

		 cWR[bankl][rank] = cycle;
		 cLastWRaddr[bankl][rank]=addr;
                  
		 dutstate.activebank[bankl+rank*dutconfig.numbank]=1'b1;
	      end
	      cLastWR = cycle;
              LastWRrankbit=addr[dutconfig.numaddr-1];
	   end

	   cmdRD, cmdRDy: begin
 	      if (modeconfig.AM && cmd == cmdRDy) begin
		 MuxXY2A(lastcmdpkt.addr, cmdpkt.addr, addr);
		 bank = lastcmdpkt.bank & dutconfig.bankaddrmask;
	         if (dutconfig.ddpranked==1) 
	           rank = lastcmdpkt.addr[dutconfig.numaddr-1];
	      end
  	      addr = addr & dutstate.bladdrmask;
		 
	      if (cmd == cmdRDy) 
		cycle=prevcycle;

	      cRD[bank][rank] = cycle;
	      cLastRD = cycle;
              LastRDrankbit=addr[dutconfig.numaddr-1];
               
	      dutstate.activebank[bank+rank*dutconfig.numbank]=1'b1;
	      if (modeconfig.DLL===1'b0)
		dutstate.qkvalid=1;
	   end

	   cmdRT, cmdRTy: begin
	      if (cmd == cmdRTy) 
		cycle=prevcycle;

	      cRT = cycle;
	   end

	   cmdLMR, cmdLMRy: begin
 	      if (modeconfig.AM && cmd == cmdLMRy) begin
		 MuxXY2A(lastcmdpkt.addr, cmdpkt.addr, addr);
		 addr = addr & dutconfig.addrmask;
		 bank = lastcmdpkt.bank & dutconfig.bankaddrmask;
	      end
              modeconfig = AddrToModeDecode(bank, addr, modeconfig);
              if (bank==0)
                mr0=addr;
              else if (bank==1)
                mr1=addr;
              else if (bank==2)
                mr2=addr;
 	      if (cmd == cmdLMRy) 
		cycle=prevcycle;

	      cLMR = cycle;
	      if ((0 == bank) &&  (modeconfig.DLL === 1'b1)) begin
		 dutstate.qkvalid=0;
	      end
	      else if ((1 == bank) &&  (modeconfig.DLLRESET === 1'b1)) begin
		 cLMRDLLRESET = cycle;
`ifdef TB
		 modeconfig.DLLRESET <= #(200ps) 1'b0;
`else
		 modeconfig.DLLRESET = 1'b0;
`endif
		 dutstate.qkvalid=0;
	      end
	      else if ((1 == bank) &&  (modeconfig.ZQe === 1'b1)) begin
`ifdef TB
		     modeconfig.ZQe <= #(200ps) 1'b0;
`else
		     modeconfig.ZQe = 1'b0;
`endif
	      end
	      // data lost when BL changes
	      if (BL !== modeconfig.BL) begin
`ifdef TB			  
		 if (TB) begin
		    $display("$ms_array_set_memory call %m at %t", $time);
		    $ms_array_set_memory(1'bx);
		 end
`endif
		 arraydata.delete();
		 arrayaccesstime.delete();

		 //update bladdrmask in dutstate
		 i=0;
		 while(dutconfig.addrmask[i] & i<MAX_ADDR_BITS) i++;
		 case (modeconfig.BL)
		   BSTL4:i=i-1;
		   BSTL8:i=i-2;
		 endcase
		 dutstate.bladdrmask='0;
		 for(j=0;j<i;j++)
		   dutstate.bladdrmask[j]=dutconfig.addrmask[j];

		 //adjust for ranked, highest bit is always CS remap bit
		 if (dutconfig.ddpranked==1) begin
		    dutstate.bladdrmask=dutstate.bladdrmask>>1;
		    dutstate.bladdrmask[dutconfig.numaddr-1]=1'b1;
		 end
	      end
	   end
	   
	   cmdREF1, cmdREF1y: begin
 	      if (modeconfig.AM && cmd == cmdREF1y) begin
		 bank = lastcmdpkt.bank & dutconfig.bankaddrmask;
	         if (dutconfig.ddpranked==1) 
	           rank = lastcmdpkt.addr[dutconfig.numaddr-1];
	      end

 	      if (cmd == cmdREF1y) 
		cycle=prevcycle;

	      if(modeconfig.REF) 
		$display("ERROR in %m at %t, updatetable, cmdREF1 not match modeconfig.REF", $time);
	      else
		cREF1[bank][rank] = cycle;
	      dutstate.refbank[bank+rank*dutconfig.numbank]=1'b1;
	      dutstate.activebank[bank+rank*dutconfig.numbank]=1'b1;
	   end
	   
	   cmdREFM, cmdREFMy: begin
 	      if (modeconfig.AM && cmdREFMy) begin
		 MuxXY2A(lastcmdpkt.addr, cmdpkt.addr, addr);
		 addr = addr & dutconfig.addrmask;
	         if (dutconfig.ddpranked==1) 
	           rank = lastcmdpkt.addr[dutconfig.numaddr-1];
	      end
 	      if (cmd == cmdREFMy) 
		cycle=prevcycle;

 	      if(modeconfig.REF) begin
		 j=0;
		 for (i=0; i< dutconfig.numbank; i++) 
		   if (addr[i]) begin
		      cREFM[i][rank] = cycle;
		      j=j+1;
		      dutstate.refbank[i+rank*dutconfig.numbank]=1'b1;
		      dutstate.activebank[i+rank*dutconfig.numbank]=1'b1;
		   end
		 if (j>1) cMMD[rank] = cycle;
		 if (j>MAX_REFM_BANKS)
		   $display("ERROR in %m at %t, updatetable, REFM banks > %0d", $time, MAX_REFM_BANKS);
	      end
	      else
		$display("ERROR in %m at %t, updatetable, cmdREFM not match modeconfig.REF", $time);
	   end
	   
           cmdNOP, cmdRDx, cmdWRx, cmdRTx, cmdLMRx, cmdREF1x, cmdREFMx: begin
              covcmd = cmd;
           end
           
	   cmdRESET: begin
`ifdef TB			  
	      if (TB) begin
		 $display("$ms_array_set_memory call %m at %t", $time);
		 $ms_array_set_memory(1'bx);
	      end
`endif
	      dutstate.qkvalid=0;
	      arraydata.delete();
	      arrayaccesstime.delete();
	   end

           default: begin
	      $display("ERROR in %m at %t, updatetable, invalid command", $time);
	   end
	   
	 endcase

	 // update dutstate
	 dutstate.cyclecount = currentcycle;
	 dutstate.cycletime = currenttime;
	 dutstate.cmd = covcmd;
	 dutstate.cmdtCK = tCK;
	 if (cmdextranops != -1) dutstate.cmdextranops = cmdextranops;
	 dutstate.cmdbank = bank;
	 dutstate.cmdaddr = addr;

	 // save time per cycle
	 cycledata[cycleind(currentcycle)].cycletime = currenttime;

	 // increase current time;
	 prevcycle=currentcycle;
	 if (currenttck!==tCK) begin
	    currenttck=tCK;
	    dutstate.qkvalid=0;
	 end
	 currenttime+=tCK;
	 currentcycle++;

	 // update latency
	 dutstate.Read_Latency = _RL();
	 dutstate.Write_Latency = _WL();
	 
	 // save last cmd for mux mode
	 lastcmdpkt=cmdpkt;
	 
	 // save time for next cycle too
	 cycledata[cycleind(currentcycle)].cycletime = currenttime;
	 
	 // clear previously used cycle data
	 clearcycledata(cycledata[cycleind((currentcycle-INITCYCLENUM+1))]);
      end
   endfunction

   function void updateAM(input amin);
      modeconfig.AM=amin;
      dutstate.Read_Latency = _RL();
      dutstate.Write_Latency = _WL();
   endfunction
   
   function void updateTMLAT3(input reg val);
      modeconfig.TMLAT3 = val;
      dutstate.Read_Latency = _RL();
      dutstate.Write_Latency = _WL();
   endfunction

   function integer _RL;
      if (modeconfig.TMLAT3)
	if (modeconfig.AM)
	  _RL = 4;
	else
	  _RL = 3;
      else
	if (modeconfig.AM)
	  _RL = modeconfig.DL+1;
	else
	  _RL = modeconfig.DL;
      if (modeconfig.DLL)
	_RL=_RL-1;
   endfunction

   function integer _WL;
      if (modeconfig.TMLAT3)
	if (modeconfig.AM)
	  _WL = 2;
	else
	  _WL = 1;
      else
	if (modeconfig.AM)
	  _WL = modeconfig.DL+2;
	else
	  _WL = modeconfig.DL+1;
   endfunction
   
   // Load needed timing
   function void updatetiming;
      input 	 UTYPE_sharedtiming ts;
      begin
	 timing = ts;
      end
   endfunction // void

   // initilized table
   function void inittable;
      begin:inittable_func
	 // states
         DefaultModeConfig;
	 // clear cycle data
	 for (int i=0; i<MAXCYCLEPIPE; i++)
	   clearcycledata(cycledata[i]);
	 
	 // something big, so all timing are satisficed
	 currenttime = INITCYCLETIME;
	 currentcycle = INITCYCLENUM;
	 
	 // clear timing
	 for (int r=0; r<2; r++) 
	   for (int i=0; i< dutconfig.numbank; i++) begin
	      cWR[i][r] = '0;
	      cWR1[i][r] = '0;
	      cWR2[i][r] = '0;
	      cRD[i][r] = '0;
	      cREF1[i][r] = '0;
	      cREFM[i][r] = '0;
	   end
	 cRT=0;
	 cLMR=0;
	 cLMRDLLRESET=0;
	 cLastRD=0;
	 cLastWR=0;
	 cMMD[0]=0;
	 cMMD[1]=0;
         scanPipeCntr=0;

	 // clear array data
	 arraydata.delete();
	 arrayaccesstime.delete();
	 
	 // dutstate init
	 dutstate.cmdstate = ST_RESET;
	 dutstate.qkvalid=0;
	 dutstate.bladdrmask=dutconfig.addrmask;

	 // TAP init
	 dutstate.curTAPstate=TAPreset;
	 nextTAPstate=TAPreset;
	 dutstate.curTAPIRstate=ID_CODE;
      end
   endfunction 

   // return at which cycle does it meet this spec from basec
   function int checkspeccycle(
    input int tCK, 
    input int basec,
    input UTYPE_checktype checktype,
    input int spec 	);
      begin: checkspeccycle_func
	 if (checktype == CKTM) // time based spec, 
	   checkspeccycle = basec + (spec+tCK-1)/tCK; //rounding
	 else // cycle based spec
	   checkspeccycle = basec + spec;
      end
   endfunction

   // check timing, how many nops are needed to meet basec + spec timing
   function int checkspec1(
    input int tCK, 
    input int targetc,
    input int basec,
    input UTYPE_checktype checktype1,
    input int spec1,
    input int curnops, 
    inout [20*8-1:0] msg,
    input [20*8-1:0] msgin
			   );
      begin: checktime1_func
	 int newcycle;
	 int nops;
	 
	 nops=curnops;
	 newcycle = checkspeccycle(tCK, basec, checktype1, spec1);
	 if (newcycle>targetc+curnops)
	   nops = newcycle-targetc;
	 
	 if (nops > curnops)
	   msg = msgin;
	 checkspec1 = nops;
      end
   endfunction

   function int checkspec2(
    input int tCK, 
    input int targetc,
    input int basec,
    input UTYPE_checktype checktype1,
    input int spec1,
    input UTYPE_checktype checktype2,
    input int spec2,
    input int curnops, 
    inout [20*8-1:0] msg, 
    input [20*8-1:0] msgin
			   );
      begin: checktime1_func
	 int newcycle;
	 int nops;
	 
	 nops=curnops;
	 newcycle = checkspeccycle(tCK, basec, checktype1, spec1);
	 newcycle = checkspeccycle(tCK, newcycle, checktype2, spec2);
	 if (newcycle>targetc+curnops)
	   nops = newcycle-targetc;
	 
	 if (nops > curnops)
	   msg = msgin;
	 checkspec2 = nops;
      end
   endfunction

   function int checkspec3(
    input int tCK, 
    input int targetc,
    input int basec,
    input UTYPE_checktype checktype1,
    input int spec1,
    input UTYPE_checktype checktype2,
    input int spec2,
    input UTYPE_checktype checktype3,
    input int spec3,
    input int curnops, 
    inout [20*8-1:0] msg,
    input [20*8-1:0] msgin
			   );
      begin: checktime1_func
	 int newcycle;
	 int nops;
	 
	 nops=curnops;
	 newcycle = checkspeccycle(tCK, basec, checktype1, spec1);
	 newcycle = checkspeccycle(tCK, newcycle, checktype2, spec2);
	 newcycle = checkspeccycle(tCK, newcycle, checktype3, spec3);
	 if (newcycle>targetc+curnops)
	   nops = newcycle-targetc;
	 
	 if (nops > curnops)
	   msg = msgin;
	 checkspec3 = nops;
      end
   endfunction

   // index to cycledata
   function int cycleind(int cycle);
      cycleind = cycle%MAXCYCLEPIPE;
   endfunction

   // current cycle time from cycle number
   function time ct(int cycle);
      if (cycle <= currentcycle)
	ct = cycledata[cycleind(cycle)].cycletime;
      else // revisit!!!, we are assuming tCK does not change here
	ct = cycledata[cycleind(currentcycle)].cycletime + (cycle-currentcycle)*currenttck ;
   endfunction

   // schedule write array 
   function void writearray(
    input [MAX_BA_BITS-1:0] bank,
    input [MAX_ADDR_BITS-1:0] addr,
    input [MAX_DQ_BITS*MAX_BURST_LEN-1:0] data,
    input [MAX_DM_BITS*MAX_BURST_LEN-1:0] dm
			    );
      begin:shcedulewritearray_func
	 reg [MAX_DQ_BITS*MAX_BURST_LEN-1:0] realdata;
	 reg [MAX_BA_BITS-1:0] 		     bankl;
	 int i,j,k;
	 time 				     tmp;
	 bank = bank & dutconfig.bankaddrmask;
	 addr = addr& dutstate.bladdrmask;

	 // combine read data with data per dm
	 for(i=0;i<numbankWRITE(modeconfig.WRITE);i++) begin
            bankl = banklWRITE(modeconfig.WRITE,i,bank);

	    readarray(bankl,addr,realdata,tmp);
	    for(j=0;j<MAX_BURST_LEN;j++)
	      for(k=0;k<MAX_DQ_BITS;k++)
		if(dutconfig.dqmask[k])
		  if (dm[j*MAX_DM_BITS+((k/9)%2)]===1'b0)
		    realdata[j*MAX_DQ_BITS+k] = data[j*MAX_DQ_BITS+k];
	    
	    arraydata[{bankl,addr}]=realdata;
	    arrayaccesstime[{bankl,addr}]=$time;
	 end
      end
   endfunction

   // array data read functions
   function void readarray(	
    input [MAX_BA_BITS-1:0] bank,
    input [MAX_ADDR_BITS-1:0] addr,
    output [MAX_DQ_BITS*MAX_BURST_LEN-1:0] data,
    output 				     time accesstime);
      begin:readarray_func
	 int count;

	 bank = bank & dutconfig.bankaddrmask;
	 addr = addr & dutstate.bladdrmask;
	 
	 if (arraydata.exists({bank,addr})) begin
	    data = arraydata[{bank,addr}];
	    accesstime = arrayaccesstime[{bank,addr}];
	 end
         else // if not exists
	    data = 'hx;
	 
      end
   endfunction

   // train data read functions
   function void readtraindata(	
    output [MAX_DQ_BITS*MAX_BURST_LEN-1:0] data);
      begin:readarray_func
	 data = 'x;
	 // data is only BL4
	 for (int i=0;i<4;i++)
	   for (int j=0;j<dutconfig.numdq;j++)
	     if (modeconfig.RTreg == 2'b01) begin
		if (j%2)
		  data[i*MAX_DQ_BITS+j] = (i+1)%2;
		else
		  data[i*MAX_DQ_BITS+j] = i%2;
	     end
	     else
	       data[i*MAX_DQ_BITS+j] = i%2;
      end
   endfunction

   function void DefaultModeConfig;
      begin
	 //reset all back to default
	 modeconfig = '0;
	 modeconfig.DL=13;
	 dutstate.Read_Latency = _RL();
	 dutstate.Write_Latency = _WL();
      end
   endfunction

   function void setmodeconfig(input UTYPE_dutmodeconfig modeconfig_in);
      modeconfig = modeconfig_in;
   endfunction

   function UTYPE_cmdpacket cmddecode(
    input RESET,
    input [1:0] CSF, 
    input WEF, 
    input REF,
    input UTYPE_cmdpacket prevcmd,
    input UTYPE_dutmodeconfig modeconfigin
				      );
      begin
	 reg csf;
	 cmddecode.cs=0;
	 case(CSF)
	   2'b00: csf=1'b0; 
	   2'b11: csf=1'b1;
	   2'b10: csf=1'b0; 
	   2'b01: begin csf=1'b0; cmddecode.cs=1; end
	 endcase
	 
	 if(modeconfigin.AM) 
	   casex({RESET, csf, WEF, REF})
	     4'b0xxx: cmddecode.cmd = cmdRESET;
	     4'b11xx: begin
		case (prevcmd.cmd)
		  cmdLMRx: cmddecode.cmd = cmdLMRy;
		  cmdWRx: cmddecode.cmd = cmdWRy;
		  cmdRDx: cmddecode.cmd = cmdRDy;
		  cmdRTx: cmddecode.cmd = cmdRTy;
		  cmdREF1x: cmddecode.cmd = cmdREF1y;
		  cmdREFMx: cmddecode.cmd = cmdREFMy;
		  default: cmddecode.cmd = cmdNOP;
		endcase
	     end
	     4'b1000: cmddecode.cmd = cmdLMRx;
	     4'b1011:
	       if (modeconfigin.RT)
		 cmddecode.cmd = cmdRTx;
	       else
		 cmddecode.cmd = cmdRDx;
	     4'b1001: cmddecode.cmd = cmdWRx;
	     4'b1010: 
	       if (modeconfigin.REF) 
		 cmddecode.cmd = cmdREFMx;
	       else
		 cmddecode.cmd = cmdREF1x;
             default: begin
		$display("ERROR in %m at %t, invalid cmd in task cmddecode, ignored", $time);
		cmddecode.cmd = cmdNOP;
             end
	   endcase 
	 else 
	   casex({RESET, csf, WEF, REF})
	     4'b0xxx: cmddecode.cmd = cmdRESET;
	     4'b11xx: cmddecode.cmd = cmdNOP;
	     4'b1000: cmddecode.cmd = cmdLMR;
	     4'b1011:
	       if (modeconfigin.RT)
		 cmddecode.cmd = cmdRT;
	       else
		 cmddecode.cmd = cmdRD;
	     4'b1001: cmddecode.cmd = cmdWR;
	     4'b1010: 
	       if (modeconfigin.REF) 
		 cmddecode.cmd = cmdREFM;
	       else
		 cmddecode.cmd = cmdREF1;
             default: begin
		$display("ERROR in %m at %t, invalid cmd in task cmddecode, ignored", $time);
		cmddecode.cmd = cmdNOP;
             end
	   endcase 
      end
   endfunction

   function UTYPE_dutmodeconfig AddrToModeDecode(
    input int ba,
    input int addr,
    input UTYPE_dutmodeconfig modeconfig_in);
      begin
         AddrToModeDecode = modeconfig_in;
         checkMRBA(ba);
         // A13 high mode retain
         if (addr[13]!==1'b1) begin
             if( ba == 2) begin
	         AddrToModeDecode.RT = MR2RT(addr);
	         AddrToModeDecode.RTreg = MR2RTREG(addr);
	         AddrToModeDecode.WRITE = MR2WRITE(addr);
             end
             else if(ba == 1) begin
	         AddrToModeDecode.DLLRESET = MR2DLLRESET(addr);
	         AddrToModeDecode.REF = MR2REF(addr);
	         AddrToModeDecode.ZQ = MR2ZQ(addr);
	         AddrToModeDecode.ZQe = MR2ZQe(addr);
	         AddrToModeDecode.BL = MR2BL(addr);
	         AddrToModeDecode.ODT = MR2ODT(addr);
	         AddrToModeDecode.ODI = MR2ODI(addr);
	         AddrToModeDecode.ODTREL = MR2ODTREL(addr);
             end
             else if(ba == 0) begin 
	         AddrToModeDecode.tRC = MR2tRC(addr);
	         AddrToModeDecode.DL = MR2DL(addr);
	         AddrToModeDecode.DLL = MR2DLL(addr);
	         AddrToModeDecode.AM = MR2AM(addr);
             end
         end
      end
   endfunction
   
   function void ModeToAddrDecode(
    input UTYPE_dutmodeconfig newmodeconfig,
    input UTYPE_dutmodeconfig prevmodeconfig,
    output reg [MAX_MR_BITS-1:0] mr0, 
    output reg [MAX_MR_BITS-1:0] mr1, 
    output reg [MAX_MR_BITS-1:0] mr2, 
    output reg [MAX_MR_REG:0] 	 mrflag);
      begin
         bit [MAX_MR_BITS-1:0] dlBits, blBits, dllBits, amBits, rtBits, trcBits, rtregBits;
         bit [MAX_MR_BITS-1:0] dllresetBits, writeBits, refBits, odtBits, odtrelBits, odiBits, zqebits, zqbits;

         // default to all 0s (but set the MR select bits)
         mr0 = MR0; mr1 = MR1; mr2 = MR2;

	 // default to no mrreg changed
	 mrflag = '0;
         
         // MR0
	 trcBits = tRC2MR(newmodeconfig.tRC);
	 dlBits = DL2MR(newmodeconfig.DL);
 	 dllBits = DLL2MR(newmodeconfig.DLL);
 	 amBits = AM2MR(newmodeconfig.AM);
	 
	 if (newmodeconfig.tRC !== prevmodeconfig.tRC) mrflag[0]=1'b1;
	 if (newmodeconfig.DL !== prevmodeconfig.DL) mrflag[0]=1'b1;
	 if (newmodeconfig.DLL !== prevmodeconfig.DLL) mrflag[0]=1'b1;
	 if (newmodeconfig.AM !== prevmodeconfig.AM) mrflag[0]=1'b1;
         
         mr0 = mr0 | dlBits | dllBits | amBits | trcBits;
         
         // MR1
	 blBits = BL2MR(newmodeconfig.BL);
	 refBits = REF2MR(newmodeconfig.REF);
 	 dllresetBits = DLLRESET2MR(newmodeconfig.DLLRESET);
	 odtBits = ODT2MR(newmodeconfig.ODT);
	 odiBits = ODI2MR(newmodeconfig.ODI);
	 odtrelBits = ODTREL2MR(newmodeconfig.ODTREL);
	 zqbits = ZQ2MR(newmodeconfig.ZQ);
	 zqebits = ZQe2MR(newmodeconfig.ZQe);
	 
	 if (newmodeconfig.BL !== prevmodeconfig.BL) mrflag[1]=1'b1;
	 if (newmodeconfig.REF !== prevmodeconfig.REF) mrflag[1]=1'b1;
	 if (newmodeconfig.DLLRESET !== prevmodeconfig.DLLRESET) mrflag[1]=1'b1;
	 if (newmodeconfig.ODT !== prevmodeconfig.ODT) mrflag[1]=1'b1;
	 if (newmodeconfig.ODTREL !== prevmodeconfig.ODTREL) mrflag[1]=1'b1;
	 if (newmodeconfig.ODI !== prevmodeconfig.ODI) mrflag[1]=1'b1;
	 if (newmodeconfig.ZQe !== prevmodeconfig.ZQe) mrflag[1]=1'b1;
	 if (newmodeconfig.ZQ !== prevmodeconfig.ZQ) mrflag[1]=1'b1;
         
         mr1 = mr1 | refBits | dllresetBits | blBits | odtrelBits | odtBits | odiBits | zqbits | zqebits;
        
         // MR2
	 rtBits = RT2MR(newmodeconfig.RT);
	 rtregBits = RTREG2MR(newmodeconfig.RTreg);
	 writeBits = WRITE2MR(newmodeconfig.WRITE);
	 
	 if (newmodeconfig.WRITE !== prevmodeconfig.WRITE) mrflag[2]=1'b1;
	 if (newmodeconfig.RT !== prevmodeconfig.RT) mrflag[2]=1'b1;
	 if (newmodeconfig.RTreg !== prevmodeconfig.RTreg) mrflag[2]=1'b1;
	 
         mr2 = mr2 | writeBits | rtBits | rtregBits;
         
      end
   endfunction

   function void MuxA2XY(
    input [MAX_ADDR_BITS-1:0]  a,
    inout [MAX_ADDR_BITS-1:0] x  , 
    inout [MAX_ADDR_BITS-1:0] y   );
      begin
	 x[0] = a[0];
	 x[3] = a[3];
	 x[4] = a[4];
	 x[5] = a[5];
	 x[8] = a[8];
	 x[9] = a[9];
	 x[10] = a[10];
	 x[13] = a[13];
	 x[14] = a[14];
	 x[17] = a[17];
	 x[18] = a[18];
	 y[3] = a[1];
	 y[4] = a[2];
	 y[5] = a[20];
	 y[8] = a[6];
	 y[9] = a[7];
	 y[10] = a[19];
	 y[13] = a[11];
	 y[14] = a[12];
	 y[17] = a[16];
	 y[18] = a[15];
	 y[5]  = a[20];

	 // for 2CS, keep CS remap bit from a
	 if (dutconfig.ddpranked==1)
	   x[dutconfig.numaddr-1] = a[dutconfig.numaddr-1];
      end
   endfunction
   
   function void MuxXY2A(
    input  [MAX_ADDR_BITS-1:0]		       x,
    input [MAX_ADDR_BITS-1:0] 		       y,
    inout [MAX_ADDR_BITS-1:0] 		       a );
      begin
	 a='0;
	 a[0]  	 = x[0]   ;
	 a[3]  	 = x[3]   ;
	 a[4]  	 = x[4]   ;
	 a[5]  	 = x[5]   ;
	 a[8]  	 = x[8]   ;
	 a[9]  	 = x[9]   ;
	 a[10]	  =x[10]    ;
	 a[13]	  =x[13]    ;
	 a[14]	  =x[14]    ;
	 a[17]	  =x[17]    ;
	 a[18]	  =x[18]    ;
	 a[1]  	 = y[3]   ;
	 a[2]  	 = y[4]   ;
	 a[20]   = y[5]   ;
	 a[6]  	 = y[8]   ;
	 a[7]  	 = y[9]   ;
	 a[19]   = y[10]   ;
	 a[11]	  =y[13]    ;
	 a[12]	  =y[14]    ;
	 a[16]	  =y[17]    ;
	 a[15]	  =y[18]    ;
	 a[20]    =y[5];

	 // for 2CS, get CS remap bit from x
	 if (dutconfig.ddpranked==1)
	   a[dutconfig.numaddr-1] = x[dutconfig.numaddr-1];
      end
   endfunction

   // TAP functions
   function void updateTAPtable (bit tms, bit tdi, output tdo, output tdo_valid);
      begin
	 // sample tdi
	 case (dutstate.curTAPstate)
	   TAPShiftDR    : begin 
	      tdo_valid = 1'b1;
	      case (TAPIR)
		ID_CODE         : begin
		   TAPID = TAPID >> 1;
		   TAPID[ID_LEN-1] = tdi;
		end
		BYPASS,HIGH_Z,CLAMP   : begin
		   TAPBYPASS = tdi;
		end
		SAMPLE_PRELOAD,EXTEST           : begin 
		   TAPBSR = TAPBSR >> 1;
		   TAPBSR[BSR_LEN-1] = tdi;
		end
		SAMPLE_PRELOAD_EXT,EXTEST_EXT           : begin 
		   TAPBSREXT = TAPBSREXT >> 1;
		   TAPBSREXT[BSR_LEN_EXT-1] = tdi;
		end
	      endcase

	   end
	   TAPShiftIR    : begin 
	      TAPIRcap = TAPIRcap >> 1;
	      TAPIRcap[INST_LEN-1] = tdi;
	   end
	 endcase
	 
	 // updates state
	 case (dutstate.curTAPstate)
	   TAPreset      : nextTAPstate = (tms) ? TAPreset       : TAPidle      ;
	   TAPidle       : nextTAPstate = (tms) ? TAPDRscan      : TAPidle      ;
	   TAPDRscan     : nextTAPstate = (tms) ? TAPIRscan      : TAPCaptureDR ;
	   TAPIRscan     : nextTAPstate = (tms) ? TAPreset       : TAPCaptureIR ;
	   TAPCaptureDR  : nextTAPstate = (tms) ? TAPExit1DR     : TAPShiftDR   ;
	   TAPCaptureIR  : nextTAPstate = (tms) ? TAPExit1IR     : TAPShiftIR   ;
	   TAPShiftDR    : nextTAPstate = (tms) ? TAPExit1DR     : TAPShiftDR   ;
	   TAPShiftIR    : nextTAPstate = (tms) ? TAPExit1IR     : TAPShiftIR   ;
	   TAPExit1DR    : nextTAPstate = (tms) ? TAPUpdateDR    : TAPPauseDR   ;
	   TAPExit1IR    : nextTAPstate = (tms) ? TAPUpdateIR    : TAPPauseIR   ;
	   TAPPauseDR    : nextTAPstate = (tms) ? TAPExit2DR     : TAPPauseDR   ;
	   TAPPauseIR    : nextTAPstate = (tms) ? TAPExit2IR     : TAPPauseIR   ;
	   TAPExit2DR    : nextTAPstate = (tms) ? TAPUpdateDR    : TAPShiftDR   ;
	   TAPExit2IR    : nextTAPstate = (tms) ? TAPUpdateIR    : TAPShiftIR   ;
	   TAPUpdateDR   : nextTAPstate = (tms) ? TAPDRscan      : TAPidle      ;
	   TAPUpdateIR   : nextTAPstate = (tms) ? TAPDRscan      : TAPidle      ;
	   default: 
	     $display("ERROR in %m at %t, updateTAPtable, unknown TAP state %0d", $time, dutstate.curTAPstate);
	 endcase
	 dutstate.curTAPstate=nextTAPstate;

	 // default no output
	 tdo_valid = 1'b0;

	 // update tdo and othere state variables
	 case (dutstate.curTAPstate)
	   TAPreset      : begin
	      dutstate.curTAPIRstate=ID_CODE;
	      TAPIR=ID_CODE;
	   end
	   TAPCaptureDR  : begin
	      case (TAPIR)
		ID_CODE                      : TAPID=dutconfig.JTAG_IDCODE;
		MR0_MR1                      : TAPID={3'b0,mr1[11:0],5'b0,mr0[11:0]};
		MR2_QED_TS                   : TAPID={22'b0, mr2[9:0]};
		BYPASS,HIGH_Z,CLAMP          : TAPBYPASS=1'b0;
		SAMPLE_PRELOAD,EXTEST        : TAPBSR=TAPBSRcap;
		SAMPLE_PRELOAD_EXT,EXTEST_EXT : TAPBSREXT=TAPBSREXTcap;
	      endcase
	   end
	   TAPCaptureIR  : begin 
	      TAPIR[1:0] = 2'b01;
	   end
	   TAPShiftDR    : begin 
	      tdo_valid = 1'b1;
	      case (TAPIR)
		ID_CODE         : begin
		   tdo = TAPID[0];
		end
		BYPASS,HIGH_Z,CLAMP   : begin
		   tdo = TAPBYPASS;
		end
		SAMPLE_PRELOAD,EXTEST           : begin 
		   tdo = TAPBSR[0];
		end
		SAMPLE_PRELOAD_EXT,EXTEST_EXT           : begin 
		   tdo = TAPBSREXT[0];
		end
	      endcase
	   end
	   TAPShiftIR    : begin 
	      tdo_valid = 1'b1;
	      tdo = TAPIRcap[0];
	   end
	     TAPUpdateDR   : begin
               if (dutstate.curTAPIRstate==EXTEST_EXT)
	         TAPBSREXTupdate = TAPBSREXT;
               else
	         TAPBSRupdate = TAPBSR;
	   end
	   TAPUpdateIR   : begin 
	      // update IR register 
	      TAPIR = TAPIRcap;
	      case (TAPIR)
		EXTEST          : dutstate.curTAPIRstate=EXTEST;
		EXTEST_EXT      : dutstate.curTAPIRstate=EXTEST_EXT;
		ID_CODE         : dutstate.curTAPIRstate=ID_CODE;
		SAMPLE_PRELOAD  : dutstate.curTAPIRstate=SAMPLE_PRELOAD;
		SAMPLE_PRELOAD_EXT  : dutstate.curTAPIRstate=SAMPLE_PRELOAD_EXT;
		CLAMP           : dutstate.curTAPIRstate=CLAMP;
		HIGH_Z          : dutstate.curTAPIRstate=HIGH_Z;
		BYPASS          : dutstate.curTAPIRstate=BYPASS;
		default         : dutstate.curTAPIRstate=TAPUNKOWN;
	      endcase
	   end
	 endcase
      end
   endfunction   
   
   function void updateTAPBSRcap(bit [BSR_LEN-1:0] in);
      TAPBSRcap = in;
   endfunction
   
   function void updateTAPBSREXTcap(bit [BSR_LEN_EXT-1:0] in);
      TAPBSREXTcap = in;
   endfunction
   
   function bit [BSR_LEN-1:0] getTAPBSR();
      getTAPBSR = TAPBSRupdate;
   endfunction

   function bit [BSR_LEN_EXT-1:0] getTAPBSREXT();
      getTAPBSREXT = TAPBSREXTupdate;
   endfunction

   // this table needs to be updated as per spec
   function int TAPpin2BSR(int pin);
      // not all pins are mapped
      TAPpin2BSR=-1;
      // 1 to 1 mapping
`ifdef rldram3_1Gb
      case(pin)
	iDQen17 : TAPpin2BSR = 1;
	iDQ17 	: TAPpin2BSR = 2;
	iDQen16 : TAPpin2BSR = 3;
	iDQ16 	: TAPpin2BSR = 4;
	iDQen15 : TAPpin2BSR = 5;
	iDQ15 	: TAPpin2BSR = 6;
	iDQen14 : TAPpin2BSR = 7;
	iDQ14 	: TAPpin2BSR = 8;
	iDQen13 : TAPpin2BSR = 9;
	iDQ13 	: TAPpin2BSR = 10;
	iDQen12 : TAPpin2BSR = 11;
	iDQ12 	: TAPpin2BSR = 12;
	iDQen11 : TAPpin2BSR = 13;
	iDQ11 	: TAPpin2BSR = 14;
	iDQen10 : TAPpin2BSR = 15;
	iDQ10 	: TAPpin2BSR = 16;
	iDQen9 	: TAPpin2BSR = 17;
	iDQ9 	: TAPpin2BSR = 18;
	iQKen1F : TAPpin2BSR = 19;
	iQK1F 	: TAPpin2BSR = 20;
	iQKen1 	: TAPpin2BSR = 21;
	iQK1 	: TAPpin2BSR = 22;
	iQVLDen0: TAPpin2BSR = 23;
	iQVLD0 	: TAPpin2BSR = 24;
	iDQen8 	: TAPpin2BSR = 25;
	iDQ8 	: TAPpin2BSR = 26;
	iDQen7 	: TAPpin2BSR = 27;
	iDQ7 	: TAPpin2BSR = 28;
	iDQen6 	: TAPpin2BSR = 29;
	iDQ6 	: TAPpin2BSR = 30;
	iDQen5 	: TAPpin2BSR = 31;
	iDQ5 	: TAPpin2BSR = 32;
	iDQen4 	: TAPpin2BSR = 33;
	iDQ4 	: TAPpin2BSR = 34;
	iDQen3 	: TAPpin2BSR = 35;
	iDQ3 	: TAPpin2BSR = 36;
	iDQen2 	: TAPpin2BSR = 37;
	iDQ2 	: TAPpin2BSR = 38;
	iDQen1 	: TAPpin2BSR = 39;
	iDQ1 	: TAPpin2BSR = 40;
	iDQen0 	: TAPpin2BSR = 41;
	iDQ0 	: TAPpin2BSR = 42;
	iQKen0F : TAPpin2BSR = 43;
	iQK0F 	: TAPpin2BSR = 44;
	iQKen0 	: TAPpin2BSR = 45;
	iQK0 	: TAPpin2BSR = 46;
	iDK1F 	: TAPpin2BSR = 47;
	iDK1 	: TAPpin2BSR = 48;
	iADDR19	: TAPpin2BSR = 49;
	iADDR14	: TAPpin2BSR = 50;
	iADDR16 : TAPpin2BSR = 51;
	iADDR15 : TAPpin2BSR = 52;
	iADDR6 	: TAPpin2BSR = 53;
	iADDR7 	: TAPpin2BSR = 54;
	iBA3 	: TAPpin2BSR = 55;
	iBA1 	: TAPpin2BSR = 56;
	iVSS0 	: TAPpin2BSR = 57;
	iVSS1 	: TAPpin2BSR = 58;
	iWEF 	: TAPpin2BSR = 59;
	iADDR0 	: TAPpin2BSR = 60;
	iADDR1 	: TAPpin2BSR = 61;
	iADDR2 	: TAPpin2BSR = 62;
        iADDR20 : TAPpin2BSR = 63;
	iADDR11 : TAPpin2BSR = 64;
	iZQ 	: TAPpin2BSR = 65;
	iDK0 	: TAPpin2BSR = 66;
	iDK0F 	: TAPpin2BSR = 67;
	iRESET 	: TAPpin2BSR = 68;
	iDM0 	: TAPpin2BSR = 69;
	iMF 	: TAPpin2BSR = 70;
	iADDR13 : TAPpin2BSR = 71;
	iADDR5 	: TAPpin2BSR = 72;
	iADDR4 	: TAPpin2BSR = 73;
	iADDR3 	: TAPpin2BSR = 74;
	iCSF0 	: TAPpin2BSR = 75;
	iVSS2 	: TAPpin2BSR = 76;
	iVSS3 	: TAPpin2BSR = 77;
	iREFF 	: TAPpin2BSR = 78;
	iCKF 	: TAPpin2BSR = 79;
	iCK 	: TAPpin2BSR = 80;
	iBA0 	: TAPpin2BSR = 81;
	iBA2 	: TAPpin2BSR = 82;
	iADDR9 	: TAPpin2BSR = 83;
	iADDR8 	: TAPpin2BSR = 84;
	iADDR18 : TAPpin2BSR = 85;
	iADDR17 : TAPpin2BSR = 86;
	iADDR12 : TAPpin2BSR = 87;
	iADDR10 : TAPpin2BSR = 88;
	iDM1 	: TAPpin2BSR = 89;
	iDQen35 : TAPpin2BSR = 90;
	iDQ35 	: TAPpin2BSR = 91;
	iDQen34 : TAPpin2BSR = 92;
	iDQ34 	: TAPpin2BSR = 93;
	iDQen33 : TAPpin2BSR = 94;
	iDQ33 	: TAPpin2BSR = 95;
	iDQen32 : TAPpin2BSR = 96;
	iDQ32 	: TAPpin2BSR = 97;
	iDQen31 : TAPpin2BSR = 98;
	iDQ31 	: TAPpin2BSR = 99;
	iDQen30 : TAPpin2BSR = 100;
	iDQ30 	: TAPpin2BSR = 101;
	iDQen29 : TAPpin2BSR = 102;
	iDQ29 	: TAPpin2BSR = 103;
	iDQen28 : TAPpin2BSR = 104;
	iDQ28 	: TAPpin2BSR = 105;
	iDQen27 : TAPpin2BSR = 106;
	iDQ27 	: TAPpin2BSR = 107;
	iQKen3F : TAPpin2BSR = 108;
	iQK3F 	: TAPpin2BSR = 109;
	iQKen3 	: TAPpin2BSR = 110;
	iQK3 	: TAPpin2BSR = 111;
	iQVLDen1: TAPpin2BSR = 112;
	iQVLD1 	: TAPpin2BSR = 113;
	iDQen26 : TAPpin2BSR = 114;
	iDQ26 	: TAPpin2BSR = 115;
	iDQen25 : TAPpin2BSR = 116;
	iDQ25 	: TAPpin2BSR = 117;
	iDQen24 : TAPpin2BSR = 118;
	iDQ24 	: TAPpin2BSR = 119;
	iDQen23 : TAPpin2BSR = 120;
	iDQ23 	: TAPpin2BSR = 121;
	iDQen22 : TAPpin2BSR = 122;
	iDQ22 	: TAPpin2BSR = 123;
	iDQen21 : TAPpin2BSR = 124;
	iDQ21 	: TAPpin2BSR = 125;
	iDQen20 : TAPpin2BSR = 126;
	iDQ20 	: TAPpin2BSR = 127;
	iDQen19 : TAPpin2BSR = 128;
	iDQ19 	: TAPpin2BSR = 129;
	iDQen18 : TAPpin2BSR = 130;
	iDQ18 	: TAPpin2BSR = 131;
	iQKen2F	: TAPpin2BSR = 132;
	iQK2F 	: TAPpin2BSR = 133;
	iQKen2 	: TAPpin2BSR = 134;
	iQK2 	: TAPpin2BSR = 135;

	//iCSF1	: TAPpin2BSR = 51;
      endcase
`else
      case(pin)
	iDQen17 : TAPpin2BSR = 1;
	iDQ17 	: TAPpin2BSR = 2;
	iDQen16 : TAPpin2BSR = 3;
	iDQ16 	: TAPpin2BSR = 4;
	iDQen15 : TAPpin2BSR = 5;
	iDQ15 	: TAPpin2BSR = 6;
	iDQen14 : TAPpin2BSR = 7;
	iDQ14 	: TAPpin2BSR = 8;
	iDQen13 : TAPpin2BSR = 9;
	iDQ13 	: TAPpin2BSR = 10;
	iDQen12 : TAPpin2BSR = 11;
	iDQ12 	: TAPpin2BSR = 12;
	iDQen11 : TAPpin2BSR = 13;
	iDQ11 	: TAPpin2BSR = 14;
	iDQen10 : TAPpin2BSR = 15;
	iDQ10 	: TAPpin2BSR = 16;
	iDQen9 	: TAPpin2BSR = 17;
	iDQ9 	: TAPpin2BSR = 18;
	iQK1F 	: TAPpin2BSR = 19;
	iQK1 	: TAPpin2BSR = 20;
	iQVLD0 	: TAPpin2BSR = 21;
	iDQen8 	: TAPpin2BSR = 22;
	iDQ8 	: TAPpin2BSR = 23;
	iDQen7 	: TAPpin2BSR = 24;
	iDQ7 	: TAPpin2BSR = 25;
	iDQen6 	: TAPpin2BSR = 26;
	iDQ6 	: TAPpin2BSR = 27;
	iDQen5 	: TAPpin2BSR = 28;
	iDQ5 	: TAPpin2BSR = 29;
	iDQen4 	: TAPpin2BSR = 30;
	iDQ4 	: TAPpin2BSR = 31;
	iDQen3 	: TAPpin2BSR = 32;
	iDQ3 	: TAPpin2BSR = 33;
	iDQen2 	: TAPpin2BSR = 34;
	iDQ2 	: TAPpin2BSR = 35;
	iDQen1 	: TAPpin2BSR = 36;
	iDQ1 	: TAPpin2BSR = 37;
	iDQen0 	: TAPpin2BSR = 38;
	iDQ0 	: TAPpin2BSR = 39;
	iQK0F 	: TAPpin2BSR = 40;
	iQK0 	: TAPpin2BSR = 41;
	iDK1F 	: TAPpin2BSR = 42;
	iDK1 	: TAPpin2BSR = 43;
	iADDR19	: TAPpin2BSR = 44;
	iADDR16 : TAPpin2BSR = 45;
	iADDR15 : TAPpin2BSR = 46;
	iADDR6 	: TAPpin2BSR = 47;
	iADDR7 	: TAPpin2BSR = 48;
	iADDR1 	: TAPpin2BSR = 49;
	iADDR2 	: TAPpin2BSR = 50;
	iCSF1	: TAPpin2BSR = 51;
	iADDR11 : TAPpin2BSR = 52;
	iZQ 	: TAPpin2BSR = 53;
	iDK0 	: TAPpin2BSR = 54;
	iDK0F 	: TAPpin2BSR = 55;
	iRESET 	: TAPpin2BSR = 56;
	iDM0 	: TAPpin2BSR = 57;
	iMF 	: TAPpin2BSR = 58;
	iADDR13 : TAPpin2BSR = 59;
	iADDR5 	: TAPpin2BSR = 60;
	iADDR4 	: TAPpin2BSR = 61;
	iADDR3 	: TAPpin2BSR = 62;
	iADDR0 	: TAPpin2BSR = 63;
	iCSF0 	: TAPpin2BSR = 64;
	iWEF 	: TAPpin2BSR = 65;
	iREFF 	: TAPpin2BSR = 66;
	iCKF 	: TAPpin2BSR = 67;
	iCK 	: TAPpin2BSR = 68;
	iBA1 	: TAPpin2BSR = 69;
	iBA0 	: TAPpin2BSR = 70;
	iBA3 	: TAPpin2BSR = 71;
	iBA2 	: TAPpin2BSR = 72;
	iADDR9 	: TAPpin2BSR = 73;
	iADDR8 	: TAPpin2BSR = 74;
	iADDR18 : TAPpin2BSR = 75;
	iADDR17 : TAPpin2BSR = 76;
	iADDR14 : TAPpin2BSR = 77;
	iADDR12 : TAPpin2BSR = 78;
	iADDR10 : TAPpin2BSR = 79;
	iDM1 	: TAPpin2BSR = 80;
	iDQen35 : TAPpin2BSR = 81;
	iDQ35 	: TAPpin2BSR = 82;
	iDQen34 : TAPpin2BSR = 83;
	iDQ34 	: TAPpin2BSR = 84;
	iDQen33 : TAPpin2BSR = 85;
	iDQ33 	: TAPpin2BSR = 86;
	iDQen32 : TAPpin2BSR = 87;
	iDQ32 	: TAPpin2BSR = 88;
	iDQen31 : TAPpin2BSR = 89;
	iDQ31 	: TAPpin2BSR = 90;
	iDQen30 : TAPpin2BSR = 91;
	iDQ30 	: TAPpin2BSR = 92;
	iDQen29 : TAPpin2BSR = 93;
	iDQ29 	: TAPpin2BSR = 94;
	iDQen28 : TAPpin2BSR = 95;
	iDQ28 	: TAPpin2BSR = 96;
	iDQen27 : TAPpin2BSR = 97;
	iDQ27 	: TAPpin2BSR = 98;
	iQK3F 	: TAPpin2BSR = 99;
	iQK3 	: TAPpin2BSR = 100;
	iQVLD1 	: TAPpin2BSR = 101;
	iDQen26 : TAPpin2BSR = 102;
	iDQ26 	: TAPpin2BSR = 103;
	iDQen25 : TAPpin2BSR = 104;
	iDQ25 	: TAPpin2BSR = 105;
	iDQen24 : TAPpin2BSR = 106;
	iDQ24 	: TAPpin2BSR = 107;
	iDQen23 : TAPpin2BSR = 108;
	iDQ23 	: TAPpin2BSR = 109;
	iDQen22 : TAPpin2BSR = 110;
	iDQ22 	: TAPpin2BSR = 111;
	iDQen21 : TAPpin2BSR = 112;
	iDQ21 	: TAPpin2BSR = 113;
	iDQen20 : TAPpin2BSR = 114;
	iDQ20 	: TAPpin2BSR = 115;
	iDQen19 : TAPpin2BSR = 116;
	iDQ19 	: TAPpin2BSR = 117;
	iDQen18 : TAPpin2BSR = 118;
	iDQ18 	: TAPpin2BSR = 119;
	iQK2F 	: TAPpin2BSR = 120;
	iQK2 	: TAPpin2BSR = 121;
      endcase
`endif
      // array starts from 0; while documentation starts from 1
      TAPpin2BSR = TAPpin2BSR-1;
   endfunction

   function int TAPBSR2pin(int bsr);
      TAPBSR2pin=-1;
      for (int i=0; i<iMAXPINS; i++)
	if (TAPpin2BSR(i) == bsr)
	  TAPBSR2pin=i;
   endfunction
   
   // this table needs to be updated as per spec
   function int TAPpin2BSREXT(int pin);
      // not all pins are mapped
      TAPpin2BSREXT=-1;
      // 1 to 1 mapping
`ifdef rldram3_1Gb
      case(pin)
	iDQen17 : TAPpin2BSREXT = 1;
	iDQ17 	: TAPpin2BSREXT = 2;
       iDQ35+500: TAPpin2BSREXT = 3;   
	iDQen16 : TAPpin2BSREXT = 4;   
	iDQ16 	: TAPpin2BSREXT = 5;   
       iDQ34+500: TAPpin2BSREXT = 6;   
	iDQen15 : TAPpin2BSREXT = 7;   
	iDQ15 	: TAPpin2BSREXT = 8;   
       iDQ33+500: TAPpin2BSREXT = 9;   
	iDQen14 : TAPpin2BSREXT = 10;  
	iDQ14 	: TAPpin2BSREXT = 11;  
       iDQ32+500: TAPpin2BSREXT = 12;  
	iDQen13 : TAPpin2BSREXT = 13;  
	iDQ13 	: TAPpin2BSREXT = 14;  
       iDQ31+500: TAPpin2BSREXT = 15;  
	iDQen12 : TAPpin2BSREXT = 16;  
	iDQ12 	: TAPpin2BSREXT = 17;  
       iDQ30+500: TAPpin2BSREXT = 18;  
	iDQen11 : TAPpin2BSREXT = 19;  
	iDQ11 	: TAPpin2BSREXT = 20;  
       iDQ29+500: TAPpin2BSREXT = 21;  
	iDQen10 : TAPpin2BSREXT = 22;  
	iDQ10 	: TAPpin2BSREXT = 23;  
       iDQ28+500: TAPpin2BSREXT = 24;  
	iDQen9 	: TAPpin2BSREXT = 25;  
	iDQ9 	: TAPpin2BSREXT = 26;  
       iDQ27+500: TAPpin2BSREXT = 27;  
	iQKen1F : TAPpin2BSREXT = 28;  
	iQK1F 	: TAPpin2BSREXT = 29;  
       iQK2F+500: TAPpin2BSREXT = 30;  
	iQKen1 	: TAPpin2BSREXT = 31;  
	iQK1 	: TAPpin2BSREXT = 32;  
       iQK3 +500: TAPpin2BSREXT = 33;  
	iQVLDen0: TAPpin2BSREXT = 34;  
	iQVLD0 	: TAPpin2BSREXT = 35;  
      iQVLD1+500: TAPpin2BSREXT = 36;  
	iDQen8 	: TAPpin2BSREXT = 37;  
	iDQ8 	: TAPpin2BSREXT = 38;  
       iDQ26+500: TAPpin2BSREXT = 39;  
	iDQen7 	: TAPpin2BSREXT = 40;  
	iDQ7 	: TAPpin2BSREXT = 41;  
       iDQ25+500: TAPpin2BSREXT = 42;  
	iDQen6 	: TAPpin2BSREXT = 43;  
	iDQ6 	: TAPpin2BSREXT = 44;  
       iDQ24+500: TAPpin2BSREXT = 45;  
	iDQen5 	: TAPpin2BSREXT = 46;  
	iDQ5 	: TAPpin2BSREXT = 47;  
       iDQ23+500: TAPpin2BSREXT = 48;  
	iDQen4 	: TAPpin2BSREXT = 49;  
	iDQ4 	: TAPpin2BSREXT = 50;  
       iDQ22+500: TAPpin2BSREXT = 51;  
	iDQen3 	: TAPpin2BSREXT = 52;  
	iDQ3 	: TAPpin2BSREXT = 53;  
       iDQ21+500: TAPpin2BSREXT = 54;  
	iDQen2 	: TAPpin2BSREXT = 55;  
	iDQ2 	: TAPpin2BSREXT = 56;  
       iDQ20+500: TAPpin2BSREXT = 57;  
	iDQen1 	: TAPpin2BSREXT = 58;  
	iDQ1 	: TAPpin2BSREXT = 59;  
       iDQ19+500: TAPpin2BSREXT = 60;  
	iDQen0 	: TAPpin2BSREXT = 61;  
	iDQ0 	: TAPpin2BSREXT = 62;  
       iDQ18+500: TAPpin2BSREXT = 63;  
	iQKen0F : TAPpin2BSREXT = 64;  
	iQK0F 	: TAPpin2BSREXT = 65;  
       iQK2F+500: TAPpin2BSREXT = 66;  
	iQKen0 	: TAPpin2BSREXT = 67;  
	iQK0 	: TAPpin2BSREXT = 68;  
       iQK2 +500: TAPpin2BSREXT = 69;  
	iDK1F 	: TAPpin2BSREXT = 70;  
       iDK1F+500: TAPpin2BSREXT = 71;  
       iDK1 +500: TAPpin2BSREXT = 72;  
	iDK1 	: TAPpin2BSREXT = 73;  
	iADDR19	: TAPpin2BSREXT = 74;  
	iADDR14	: TAPpin2BSREXT = 75;  
	iADDR16 : TAPpin2BSREXT = 76;  
	iADDR15 : TAPpin2BSREXT = 77;  
	iADDR6 	: TAPpin2BSREXT = 78;  
	iADDR7 	: TAPpin2BSREXT = 79;  
	iBA3 	: TAPpin2BSREXT = 80;  
	iBA1 	: TAPpin2BSREXT = 81;  
	iVSS0 	: TAPpin2BSREXT = 82;  
	iVSS1 	: TAPpin2BSREXT = 83;  
	iWEF 	: TAPpin2BSREXT = 84;  
	iADDR0 	: TAPpin2BSREXT = 85;  
	iADDR1 	: TAPpin2BSREXT = 86;  
	iADDR2 	: TAPpin2BSREXT = 87;  
        iADDR20 : TAPpin2BSREXT = 88;  
	iADDR11 : TAPpin2BSREXT = 89;  
	iZQ 	: TAPpin2BSREXT = 90;  
	iDK0 	: TAPpin2BSREXT = 91;  
	iDK0F 	: TAPpin2BSREXT = 92;  
       iDK0F+500: TAPpin2BSREXT = 93;  
       iDK0 +500: TAPpin2BSREXT = 94;  
	iRESET 	: TAPpin2BSREXT = 95;  
	iDM0 	: TAPpin2BSREXT = 96;  
       iDM0 +500: TAPpin2BSREXT = 97;  
	iMF 	: TAPpin2BSREXT = 98;  
	iADDR13 : TAPpin2BSREXT = 99;  
	iADDR5 	: TAPpin2BSREXT = 100; 
	iADDR4 	: TAPpin2BSREXT = 101; 
	iADDR3 	: TAPpin2BSREXT = 102; 
	iCSF0 	: TAPpin2BSREXT = 103; 
	iVSS2 	: TAPpin2BSREXT = 104; 
	iVSS3 	: TAPpin2BSREXT = 105; 
	iREFF 	: TAPpin2BSREXT = 106; 
	iCKF 	: TAPpin2BSREXT = 107; 
	iCK 	: TAPpin2BSREXT = 108; 
	iBA0 	: TAPpin2BSREXT = 109; 
	iBA2 	: TAPpin2BSREXT = 110; 
	iADDR9 	: TAPpin2BSREXT = 111; 
	iADDR8 	: TAPpin2BSREXT = 112; 
	iADDR18 : TAPpin2BSREXT = 113; 
	iADDR17 : TAPpin2BSREXT = 114; 
	iADDR12 : TAPpin2BSREXT = 115; 
	iADDR10 : TAPpin2BSREXT = 116; 
	iDM1 	: TAPpin2BSREXT = 117; 
       iDM1 +500: TAPpin2BSREXT = 118; 
	iDQen35 : TAPpin2BSREXT = 119; 
	iDQ35 	: TAPpin2BSREXT = 120; 
       iDQ17+500: TAPpin2BSREXT = 121; 
	iDQen34 : TAPpin2BSREXT = 122; 
	iDQ34 	: TAPpin2BSREXT = 123; 
       iDQ16+500: TAPpin2BSREXT = 124; 
	iDQen33 : TAPpin2BSREXT = 125; 
	iDQ33 	: TAPpin2BSREXT = 126; 
       iDQ15+500: TAPpin2BSREXT = 127; 
	iDQen32 : TAPpin2BSREXT = 128; 
	iDQ32 	: TAPpin2BSREXT = 129; 
       iDQ14+500: TAPpin2BSREXT = 130; 
	iDQen31 : TAPpin2BSREXT = 131; 
	iDQ31 	: TAPpin2BSREXT = 132; 
       iDQ13+500: TAPpin2BSREXT = 133; 
	iDQen30 : TAPpin2BSREXT = 134; 
	iDQ30 	: TAPpin2BSREXT = 135; 
       iDQ12+500: TAPpin2BSREXT = 136; 
	iDQen29 : TAPpin2BSREXT = 137; 
	iDQ29 	: TAPpin2BSREXT = 138; 
       iDQ11+500: TAPpin2BSREXT = 139; 
	iDQen28 : TAPpin2BSREXT = 140; 
	iDQ28 	: TAPpin2BSREXT = 141; 
       iDQ10+500: TAPpin2BSREXT = 142; 
	iDQen27 : TAPpin2BSREXT = 143; 
	iDQ27 	: TAPpin2BSREXT = 144; 
       iDQ9 +500: TAPpin2BSREXT = 145; 
	iQKen3F : TAPpin2BSREXT = 146; 
	iQK3F 	: TAPpin2BSREXT = 147; 
       iQK1F+500: TAPpin2BSREXT = 148; 
	iQKen3 	: TAPpin2BSREXT = 149; 
	iQK3 	: TAPpin2BSREXT = 150; 
       iQK1 +500: TAPpin2BSREXT = 151; 
	iQVLDen1: TAPpin2BSREXT = 152; 
	iQVLD1 	: TAPpin2BSREXT = 153; 
      iQVLD0+500: TAPpin2BSREXT = 154; 
	iDQen26 : TAPpin2BSREXT = 155; 
	iDQ26 	: TAPpin2BSREXT = 156; 
        iDQ8+500: TAPpin2BSREXT = 157; 
	iDQen25 : TAPpin2BSREXT = 158; 
	iDQ25 	: TAPpin2BSREXT = 159; 
        iDQ7+500: TAPpin2BSREXT = 160; 
	iDQen24 : TAPpin2BSREXT = 161; 
	iDQ24 	: TAPpin2BSREXT = 162; 
        iDQ6+500: TAPpin2BSREXT = 163; 
	iDQen23 : TAPpin2BSREXT = 164; 
	iDQ23 	: TAPpin2BSREXT = 165; 
        iDQ5+500: TAPpin2BSREXT = 166; 
	iDQen22 : TAPpin2BSREXT = 167; 
	iDQ22 	: TAPpin2BSREXT = 168; 
        iDQ4+500: TAPpin2BSREXT = 169; 
	iDQen21 : TAPpin2BSREXT = 170; 
	iDQ21 	: TAPpin2BSREXT = 171; 
        iDQ3+500: TAPpin2BSREXT = 172; 
	iDQen20 : TAPpin2BSREXT = 173; 
	iDQ20 	: TAPpin2BSREXT = 174; 
        iDQ2+500: TAPpin2BSREXT = 175; 
	iDQen19 : TAPpin2BSREXT = 176; 
	iDQ19 	: TAPpin2BSREXT = 177; 
        iDQ1+500: TAPpin2BSREXT = 178; 
	iDQen18 : TAPpin2BSREXT = 179; 
	iDQ18 	: TAPpin2BSREXT = 180; 
        iDQ0+500: TAPpin2BSREXT = 181; 
	iQKen2F	: TAPpin2BSREXT = 182; 
	iQK2F 	: TAPpin2BSREXT = 183; 
       iQK0F+500: TAPpin2BSREXT = 184; 
	iQKen2 	: TAPpin2BSREXT = 185; 
	iQK2 	: TAPpin2BSREXT = 186; 
        iQK0+500: TAPpin2BSREXT = 187; 

	//iCSF1	: TAPpin2BSREXT = 51;
      endcase
`endif
      // array starts from 0; while documentation starts from 1
      TAPpin2BSREXT = TAPpin2BSREXT-1;
   endfunction

   function int TAPBSREXT2pin(int bsr);
      TAPBSREXT2pin=-1;
      for (int i=0; i<iMAXPINS+500; i++)
	if (TAPpin2BSREXT(i) == bsr)
	  TAPBSREXT2pin=i;
   endfunction
   
endmodule

